var FS = require('./../FamilySearch'),
    utils = require('./../utils'),
    maybe = utils.maybe;
    
/**
 * @ngdoc overview
 * @name name
 * @description
 * Name
 */

/**
 * @ngdoc function
 * @name name.types:constructor.Name
 * @description
 *
 * Name
 * 
 * @param {FamilySearch} client FamilySearch sdk client
 * @param {Object|String=} data either a fullText string or an object with optional attributes
 * {type, givenName, surname, prefix, suffix, fullText, preferred, changeMessage}
 */
var Name = FS.Name = function(client, data) {
  FS.BaseClass.call(this, client, data);
  
  if(data){
    if(utils.isString(data)){
      this.data = {};
      this.setFullText(data);
    } else {
      if (data.type) {
        //noinspection JSUnresolvedFunction
        this.setType(data.type);
      }
      if (data.givenName) {
        //noinspection JSUnresolvedFunction
        this.setGivenName(data.givenName);
        delete data.givenName;
      }
      if (data.surname) {
        //noinspection JSUnresolvedFunction
        this.setSurname(data.surname);
        delete data.surname;
      }
      if (data.prefix) {
        //noinspection JSUnresolvedFunction
        this.setPrefix(data.prefix);
        delete data.prefix;
      }
      if (data.suffix) {
        //noinspection JSUnresolvedFunction
        this.setSuffix(data.suffix);
        delete data.suffix;
      }
      if (data.fullText) {
        //noinspection JSUnresolvedFunction
        this.setFullText(data.fullText);
        delete data.fullText;
      }
      //noinspection JSUnresolvedFunction
      this.setPreferred(!!data.preferred);
      if (data.changeMessage) {
        //noinspection JSUnresolvedFunction
        this.setChangeMessage(data.changeMessage);
        delete data.changeMessage;
      }
      if (data.attribution && !(data.attribution instanceof FS.Attribution)) {
        this.attribution = client.createAttribution(data.attribution);
      }
    }
  }
};

/**
 * @ngdoc function
 * @name name.functions:createName
 * @param {Object} data [Name](https://familysearch.org/developers/docs/api/gx/Name_json) data
 * @return {Object} {@link name.types:constructor.Name Name}
 * @description Create a {@link name.types:constructor.Name Name} object. Use this method instead of calling the constructor directly.
 */
FS.prototype.createName = function(data){
  return new Name(this, data);
};

// return the i'th name form; add it if it doesn't exist
function ensureNameForm(name, i) {
  var pos = i || 0; // just to be clear
  if (!utils.isArray(name.data.nameForms)) {
    name.data.nameForms = [];
  }
  while (pos >= name.data.nameForms.length) {
    name.data.nameForms.push({});
  }
  return name.data.nameForms[pos];
}

Name.prototype = utils.extend(Object.create(FS.BaseClass.prototype), {
  
  constructor: Name,
  
  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getId
   * @methodOf name.types:constructor.Name
   * @return {String} Id of the name
   */

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getType
   * @methodOf name.types:constructor.Name
   * @return {String} http://gedcomx.org/BirthName, etc.
   */
  getType: function(){ return this.data.type; },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#isPreferred
   * @methodOf name.types:constructor.Name
   * @return {Boolean} true if this name is preferred
   */
  isPreferred: function(){ return this.data.preferred; },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getAttribution
   * @methodOf name.types:constructor.Name
   * @returns {Attribution} {@link attribution.types:constructor.Attribution Attribution} object
   */

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getNameFormsCount
   * @methodOf name.types:constructor.Name

   * @return {Number} get the number of name forms
   */
  getNameFormsCount: function() { return this.data.nameForms ? this.data.nameForms.length : 0; },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getNameForm
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {Number} get the `i`'th name form: each name form has `lang`, `fullText`, and `parts` properties
   */
  getNameForm: function(i) { return maybe(this.data.nameForms)[i || 0]; },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getFullText
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the full text of the `i`'th name form
   */
  getFullText: function(i) { return maybe(this.getNameForm(i)).fullText; },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getLang
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the language of the `i`'th name form
   */
  getLang: function(i) { return maybe(this.getNameForm(i)).lang; },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getNamePart
   * @methodOf name.types:constructor.Name

   * @description you can call getGivenName, getSurname, getPrefix, or getSuffix instead of this function
   * @param {String} type http://gedcomx.org/Given or http://gedcomx.org/Surname
   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the specified part of the `i`'th name form
   */
  getNamePart: function(type, i) {
    return maybe(utils.find(maybe(this.getNameForm(i)).parts, {type: type})).value;
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getGivenName
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the given part of the `i`'th name form
   */
  getGivenName: function(i) {
    return this.getNamePart('http://gedcomx.org/Given', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getSurname
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the surname part of the `i`'th name form
   */
  getSurname: function(i) {
    return this.getNamePart('http://gedcomx.org/Surname', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getPrefix
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the prefix part of the `i`'th name form
   */
  getPrefix: function(i) {
    return this.getNamePart('http://gedcomx.org/Prefix', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#getSuffix
   * @methodOf name.types:constructor.Name

   * @param {Number=} i name form to read; defaults to 0
   * @return {String} get the suffix part of the `i`'th name form
   */
  getSuffix: function(i) {
    return this.getNamePart('http://gedcomx.org/Suffix', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setType
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   * @param {String} type e.g., http://gedcomx.org/BirthName
   * @return {Name} this name
   */
  setType: function(type) {
    this.changed = true;
    if (!!type) {
      this.data.type = type;
    }
    else {
      delete this.data.type;
    }
    //noinspection JSValidateTypes
    return this;
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setPreferred
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   *
   * __NOTE__: the preferred name flag can be set only when the person is initially created; after that it is read-only
   * @param {boolean} isPreferred true if preferred
   * @return {Name} this name
   */
  setPreferred: function(isPreferred) {
    this.changed = true;
    this.data.preferred = isPreferred;
    //noinspection JSValidateTypes
    return this;
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setFullText
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   * @param {String} fullText value
   * @param {Number=} i name form to set; defaults to 0
   * @return {Name} this name
   */
  setFullText: function(fullText, i) {
    this.changed = true;
    var nameForm = ensureNameForm(this, i);
    if (!!fullText) {
      nameForm.fullText = fullText;
    }
    else {
      delete nameForm.fullText;
    }
    //noinspection JSValidateTypes
    return this;
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setNamePart
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value;
   * you can call setGivenName, setSurname, setPrefix, and setSuffix instead of this function
   * @param {String} name value
   * @param {String} type http://gedcomx.org/Given or http://gedcomx.org/Surname
   * @param {Number=} i name form to set; defaults to 0
   * @return {Name} this name
   */
  setNamePart: function(name, type, i) {
    this.changed = true;
    var nameForm = ensureNameForm(this, i);
    if (!utils.isArray(nameForm.parts)) {
      nameForm.parts = [];
    }
    var part = utils.find(nameForm.parts, {type: type});
    if (!!name) {
      if (!part) {
        part = {type: type};
        nameForm.parts.push(part);
      }
      part.value = name;
    }
    else if (!!part) {
      nameForm.parts.splice(nameForm.parts.indexOf(part), 1);
    }
    //noinspection JSValidateTypes
    return this;
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setGivenName
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   * @param {String} givenName value
   * @param {Number=} i name form to set; defaults to 0
   * @return {Name} this name
   */
  setGivenName: function(givenName, i) {
    return this.setNamePart(givenName, 'http://gedcomx.org/Given', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setSurname
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   * @param {String} surname value
   * @param {Number=} i name form to set; defaults to 0
   * @return {Name} this name
   */
  setSurname: function(surname, i) {
    return this.setNamePart(surname, 'http://gedcomx.org/Surname', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setPrefix
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   * @param {String} prefix value
   * @param {Number=} i name form to set; defaults to 0
   * @return {Name} this name
   */
  setPrefix: function(prefix, i) {
    return this.setNamePart(prefix, 'http://gedcomx.org/Prefix', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setSuffix
   * @methodOf name.types:constructor.Name

   * @description sets the changed flag as well as the value
   * @param {String} suffix value
   * @param {Number=} i name form to set; defaults to 0
   * @return {Name} this name
   */
  setSuffix: function(suffix, i) {
    return this.setNamePart(suffix, 'http://gedcomx.org/Suffix', i);
  },

  /**
   * @ngdoc function
   * @name name.types:constructor.Name#setChangeMessage
   * @methodOf name.types:constructor.Name

   * @description sets the changeMessage used to update the name
   * @param {String} changeMessage change message
   * @return {Name} this name
   */
  setChangeMessage: function(changeMessage) {
    this.setAttribution(this.client.createAttribution(changeMessage));
    //noinspection JSValidateTypes
    return this;
  }
});
