var FS = require('./../FamilySearch'),
    utils = require('./../utils'),
    maybe = utils.maybe;

/**
 * @ngdoc function
 * @name notes.types:constructor.Note
 * @description
 *
 * Note
 * 
 * To create a new note, you must set subject and text.
 * 
 * @param {FamilySearch} client FamilySearch sdk client
 * @param {Object=} data an object with optional attributes {subject, text, attribution}
 */
var Note = FS.Note = function(client, data) {
  FS.BaseClass.call(this, client, data);
  if(this.attribution && !(this.attribution instanceof FS.Attribution)){
    this.attribution = client.createAttribution(this.attribution);
  }
};

/**
 * @ngdoc function
 * @name notes.functions:createNote
 * @param {Object} data [Note](https://familysearch.org/developers/docs/api/gx/Note_json) data
 * @return {Object} {@link notes.types:constructor.Note Note}
 * @description Create a {@link notes.types:constructor.Note Note} object. Use this method instead of calling the constructor directly.
 */
FS.prototype.createNote = function(data){
  return new Note(this, data);
};

Note.prototype = utils.extend(Object.create(FS.BaseClass.prototype), {
  
  constructor: Note,
  
  /**
   * @ngdoc function
   * @name notes.types:constructor.Note#getId
   * @methodOf notes.types:constructor.Note
   * @return {String} Id of the note
   */

  /**
   * @ngdoc function
   * @name notes.types:constructor.Note#getSubject
   * @methodOf notes.types:constructor.Note
   * @return {String} subject / title of the note
   */
  getSubject: function() { return this.data.subject; },

  /**
   * @ngdoc function
   * @name notes.types:constructor.Note#getText
   * @methodOf notes.types:constructor.Note
   * @return {String} text of the note
   */
  getText: function() { return this.data.text; },

  /**
   * @ngdoc function
   * @name notes.types:constructor.Note#getNoteUrl
   * @methodOf notes.types:constructor.Note

   * @return {String} note URL (without the access token)
   */
  getNoteUrl: function() { return this.helpers.removeAccessToken(maybe(this.getLink('note')).href); },

  /**
   * @ngdoc function
   * @name notes.types:constructor.Note#save
   * @methodOf notes.types:constructor.Note

   * @description
   * Create a new note (if this note does not have an id) or update the existing note.
   *
   *
   * @param {string} url url of the notes list endpoint; only necessary when creating a new note
   * @param {string} changeMessage change message
   * @return {Object} promise for the response
   * and if refresh is true, after the note has been read.
   */
  save: function(url, changeMessage) {
    var self = this;
    if(!url){
      url = self.getNoteUrl();
    }
    var headers = {};
    var entityType = self.helpers.getEntityType(url);
    if (entityType === 'childAndParentsRelationships') {
      headers['Content-Type'] = 'application/x-fs-v1+json';
    }
    var payload = {};
    payload[entityType] = [ { notes: [ self ] } ];
    if (changeMessage) {
      payload[entityType][0].attribution = self.client.createAttribution(changeMessage);
    }
    return self.plumbing.post(url, payload, headers).then(function(response){
      self.updateFromResponse(response, 'note');
      return response;
    });
  },

  /**
   * @ngdoc function
   * @name notes.types:constructor.Note#delete
   * @methodOf notes.types:constructor.Note

   * @description delete this note
   * @param {string=} changeMessage change message
   * @return {Object} promise for the response
   */
  delete: function(changeMessage) {
    return this.client.deleteNote(this.getNoteUrl(), changeMessage);
  }

});